require 'fileutils'
require 'find'
require 'net/http'
require 'timeout'
require 'uri'
require 'matrix'

#inherits from math.rb's Vector
class Vector
	def dotProduct(p)
		x=0
		y=1
		return self[x] * p[x] + self[y] * p[y]
	end
	def angle(v2)
		v1 = self
		dot = v1.dotProduct(v2)
		length_product = (v1.r*v2.r)
		if ((dot-length_product).abs < 0.000001)
			return 0.0
		else
			#$stderr.print(sprintf("length_product: %f dot: %f\n", length_product, dot))
			#$stderr.print(sprintf("v1: %s v2: %s v1.r: %f v2.r: %f\n", v1.to_s, v2.to_s, v1.r, v2.r))
		end
		#if (dot/length_product >= 1)
		#end
		angle_radians = Math.acos(dot/length_product)
		return (180/Math::PI)*angle_radians
	end
end


module OSMModel
def debug(string)
	#$stderr.puts string
end
module_function :debug

def hash_to_tags(h)
	s = ''
	if ! h.empty?
		s += "\n"
	end
	h.each_pair do |key, value|
		if (value.nil? || (value.to_s.length == 0))
			#$stderr.print(sprintf("hash_to_tags() bad value[%s]: '%s'\n", key, value.to_s))
			next
		end
		value.gsub(/\'/, "\\'")
		value.gsub(/\&/, "amp")
		s += "\t<tag k=\"#{key}\" v=\"#{value}\"/>\n"
	end
	s
end
module_function :hash_to_tags
$way_types = ["highway", "railway", "power", "man_made", "route"]
def tags_to_feature_type(tags)
	$way_types.each do |way_type|
		next if tags[way_type.to_s].nil?
		return way_type
	end
	return nil
end
module_function :tags_to_feature_type

$way_cache
def create_way(coalesce, segs, tags = {})
	if $way_cache.nil?
		$way_cache = Hash.new
	end
    	name = tags["name"]
 	if $way_cache[name].nil?
    		#debug("creating array for \"" + name + "\"")
    		$way_cache[name] = []
    	end
    	new_way = OSMModel::Way.new(segs, tags)
	$way_cache[name].push(new_way)
  	$way_cache[name].each do |way|
		break if ! coalesce
    		if way.combine_with(new_way)
			# "new_way" was combined into another
			# now we need to get rid of it
    			$way_cache[name].delete(new_way)
	    		new_way = way
		end
    	end
end
module_function :create_way
def get_all_ways
	return $way_cache
end
module_function :get_all_ways

$node_cache
def new_node(point, feature_type, tags)
        node = nil
	lat = point[1]
	lon = point[0]
        loop = 0;
	if $node_cache.nil?
		$node_cache = Hash.new
	end
	#stderr.print(sprintf("new_node() feature_type: '%s'\n", feature_type))
        key = lat+lon
        while $node_cache.has_key?(key)
                node = $node_cache[key]
       		if node.feature_type == feature_type &&
		   node.distance(lat,lon) < 0.0000001
			#print "found close node\n"
			if node.combine_tags(tags)
	                	#stderr.puts "cache hit"
				return node
			end
		else
			#printf "actual hash collision: %f,%f vs %f,%f\n", lat, lon, node.lat, node.lon
		end
                key += lat+lon
                loop+=1
                #stderr.puts "loop nr: #{loop}" if loop > 2
        end
        node = Node.new(point, feature_type, tags)
        while $node_cache.has_key?(key)
                key += lat+lon
        end
        $node_cache[key] = node
        node
end
module_function :new_node
$node_ids = -1
class Node
	attr	:lat
	attr	:lon
	attr_accessor :osmid
	attr_accessor :feature_type
        def initialize point, feature_type, tags={}
	      @visible="true"
	      @lat = point[1]
	      @lon = point[0]
	      @feature_type = feature_type
	      @tags = tags
	      @osmid = ($node_ids-=1)
	end
	def -(node)
		xdiff = self.lon.to_f - node.lon.to_f
		ydiff = self.lat.to_f - node.lat.to_f
		xdiff *= 100.0
		ydiff *= 100.0
		#printf "lat/lon: %f/%f\n", self.lat, self.lon
		#printf "xdiff: %f\n", xdiff
		#printf "ydiff: %f\n", ydiff
		return Vector[xdiff, ydiff]
	end
	def combine_tags(tags)
		replacements = Hash.new
		tags.each_pair do |key, value|
			next if @tags[key] == value
			if key == "tiger:tzid"
				replacements[key] = @tags[key] + ":" + value
				next
			end
			OSMModel::debug(sprintf("incompatible tag: '%s' values: '%s' '%s'\n", key, value, @tags[key]))
			return false
		end
		replacements.each_pair do |key, value|
			@tags[key] = replacements[key]
		end
		return true
	end
	def same_position_as(node)
		if node.lat == self.lat &&
		   node.lon == self.lon
			return 1
		end
		return 0
	end
	def distance(lat, lon)
		if lat == self.lat &&
		   lon == self.lon
			return 0
		end
		latdiff = (lat - self.lat).abs
		londiff = (lon - self.lon).abs
		return Math.sqrt(latdiff*latdiff + londiff*londiff)
	end
	def to_xml
		return "" if @printed
		@printed = true
		return "<node id=\'#{@osmid}\' visible=\'true\' lat=\'#{@lat}\' lon=\'#{@lon}\'>" +
			OSMModel.hash_to_tags(@tags) + "</node>\n"
	end
	def id
		@osmid
	end
	def to_s
		"lat: #{@lat} lon: #{@lon} id: #{@id}"
	end
end
 
# Create a new object
# g = Greeter.new("world")
#  
#  # Output "Hello World!"
#  g.salute
$segment_ids = -1000000;
class Segment
	attr :to
	attr :from
	attr :tags
	attr_accessor :osmid
	def initialize(from, to, tags = {})
		@from = from
		@to = to
		@tags = tags
		@printed = false
		#OSMModel::debug("created segment: " + self.to_s)
		@osmid = ($segment_ids-=1)
  	end
	def vector
		return (from - to)
	end
	def length
		return vector.r
	end
	def comes_before(seg)
		if self.to == seg.from
			return true
		end
		return false
	end
	def reverse
		tmp = @from
		@from = @to
		@to = tmp
	end
	def to_xml
		return "" if @printed
		@printed = true
		return from.to_xml() +
		       to.to_xml() + 
		"<segment from=\'#{from.id}\' to=\'#{to.id}\' id=\'#{self.id}\' visible=\'true\'>" +
		OSMModel.hash_to_tags(@tags) +
		"</segment>\n"	
	end
	def id
		@osmid
	end
	def to_s
		"segment:\{id:#{@osmid} from:#{@from.to_s} to:#{@to.to_s}\}"
	end
end
 
def create_segment(from_node, to_node, tags = {})
	return Segment.new(from_node, to_node, tags)
end

$way_ids = -2000000;
class Way
	attr :segs
	attr :tags
	attr_accessor :osmid
	def initialize(segs, tags = {})
		@osmid = ($way_ids-=1)
		@segs = []
		#stderr.printf("adding segments to new way: %d\n", @osmid)
		@tags = tags
		segs.each do |seg|
			self.add_segment(seg)
		end
		#stderr.printf("done adding segments to new way: %d\n", @osmid)
  	end
	def reverse
		@segs.each do |seg|
			seg.reverse
		end
		@segs.reverse!
	end
	def name
		return @tags["name"]
	end
	def __add_segment(newseg)
		result = false
		if @segs.empty?
			OSMModel::debug("added " + newseg.to_s  + " to empty way #{@osmid}")
			@segs.push(newseg)
			result = true
		elsif newseg.comes_before(@segs.first)
			OSMModel::debug("added "+newseg.to_s+" to head of way #{@osmid} before " + @segs.first.to_s)
			@segs.unshift(newseg)
			result = true
		elsif @segs.last.comes_before(newseg)
			OSMModel::debug("added "+newseg.to_s+" to tail of way #{@osmid} after " + @segs.last.to_s)
			@segs.push(newseg)
			result = true
		end
		return result
	end
	def add_segment(newseg)
		if __add_segment(newseg)
			return true
		end
		# try the segment in the opposite direction
		newseg = Segment.new(newseg.to, newseg.from, newseg.tags)
		return __add_segment(newseg)
	end
	def combine_with(way)
		return if way == self
		# check tag compatibility here
		if ((way.segs.first.to == @segs.first.to) ||
		    (way.segs.last.to ==  @segs.last.to))
			way.reverse
		end
		combined = false
		max_angle = 60
		if @segs.last.to == way.segs.first.from
			v1 = @segs.last.vector
			v2 = way.segs.first.vector
			angle = v1.angle(v2)
			if (angle < max_angle)
				@segs.concat(way.segs)
				combined = true
			end
		elsif way.segs.last.to == @segs.first.from
			v1 = @segs.first.vector
			v2 = way.segs.last.vector
			angle = v1.angle(v2)
			if (angle < max_angle)
				@segs = way.segs | @segs
				combined = true
			end
		end
		# if the passed in way is empty, we've succeeded
		return combined
	end
	def to_xml
		@segs.map { |seg| seg.to_xml }.join +
		"<way id=\'#{@osmid}' visible=\'true\'>" +
		@segs.map { |seg| "\n\t<seg id=\'#{seg.id}\'/>" }.join +
		OSMModel.hash_to_tags(@tags) +
		"</way>\n"
	end
	def name
		return @tags["name"]
	end
	def id
		@osmid
	end
	def verify
		return
		last_seg = nil
		@segs.each do |seg|
			if !last_seg.nil?
				if last_seg.to != seg.from
					$stderr.puts "invalid way: " + self.to_s
					exit(1)
				end
			end
			last_seg = seg
		end
	end
	def to_s
		ret = "way:\{id:#{@osmid} segments: ["
		@segs.each do |seg|
			ret += "\n   " + seg.to_s + ","
		end
		ret += "]\}"
	end
end

end
